import os
import sys
from importlib import util as _import_util
from subprocess import run


class MagicFinder:
	"""
	Contextual module finder which claims to know about packages given by user

	In the following example if the 'foo' and 'bar' are not found on `sys.path`
	the configured packages will be installed with pip.  Any pip package name
	can be given, including source and VCS URLs:

	> with MagicFinder(foo='python-foo', bar='git+https://github.com/example/pybar'):
	>     import foo.fobulator
	>     import bar
	"""

	def __init__(self, **pip_packages):
		self.packages = pip_packages

	def __enter__(self):
		sys.meta_path.append(self)

	def __exit__(self, *exc_info):
		sys.meta_path.remove(self)

	def find_spec(self, name, path, target):
		if name not in self.packages or path is not None or target is not None:
			return
		# name is popped off self.packages to prevent recursion
		return MagicSpec(name, self.packages.pop(name))


class MagicSpec:
	"""
	Module spec & loader wrapper which handles installation

	After installation the wrapper hands off to an actual loader.
	"""

	def __init__(self, name, package):
		self.name = name
		self.package = package
		self.loader = self
		self.spec = None

	def __getattr__(self, name):
		return getattr(self.spec, name)

	def create_module(self, _):
		cmd = [sys.executable, "-m", "pip", "install", self.package]
		if not os.access(sys.prefix, os.W_OK):
			cmd.insert(1, '--user')
		proc = run(cmd)
		if 0 != proc.returncode:
			raise ImportError("Unable to install {}".format(self.package))
		self.spec = _import_util.find_spec(self.name)
		if hasattr(self.spec.loader, 'create_module'):
			return self.spec.loader.create_module(self.spec)

	def exec_module(self, module):
		module.__spec__ = self.spec
		return self.spec.loader.exec_module(module)
