. ~/.shell/lib/constants.sh
. ~/.shell/lib/shared.sh

: ${SSH_BIN:=`which ssh`}; export SSH_BIN

if [ $PLATFORM = windows ]; then
	if ! has sed; then
		die "This script requires 'sed'"
	fi
	if ! has awk; then
		die "This script requires 'awk'"
	fi
fi


ssh_version() {
	eval `${SSH_BIN} -V 2>&1 | sed -n \
		's/OpenSSH[a-zA-Z_]*_\([0-9]\{1,\}\)\.\([0-9]\{1,\}\).*/SSH_MAJOR="\1" SSH_MINOR="\2"; return 0/p'`
	echo >&2 "Failed to get SSH version"
	return 1
}

make_config() {
	ssh_version || return 1

	# SSH >=7.3 does not need a filtered config
	[ $SSH_MAJOR -gt 7 ] && return
	[ $SSH_MAJOR -eq 7 ] && [ $SSH_MINOR -ge 3 ] && return

	SSH_CONFIG=$HOME/.ssh/configs/$PLATFORM.$KERNEL.$SSH_MAJOR.$SSH_MINOR.config
	export SSH_CONFIG

	if [ -f "$SSH_CONFIG" ] &&
		[ "$SSH_CONFIG" -nt ~/.ssh/config ] &&
		[ "$SSH_CONFIG" -nt ~/.shell/lib/ssh.sh ] &&
		[ "$SSH_CONFIG" -nt $0 ] &&
		(! [ -r ~/.ssh/config.local ] || [ "$SSH_CONFIG" -nt ~/.ssh/config.local ])
	then
		return
	fi

	mkdir -p `dirname $SSH_CONFIG`

	cat ~/.ssh/config ~/.ssh/config.local 2>/dev/null |
		awk >"$SSH_CONFIG" "
			function remove_line() { sub(/^/, \"#\"); }
			/^$/ || /^#[^@]/ { print \$0; next }

			`rule_match`
			`rule_unix_sockets`
			`rule_mux`
			`rule_mux_persist`
			`rule_canonisation`
			`rule_percent_C`
			`rule_include`
			`rule_jump`

			1
		"

	# Ensure the GPG agent is running if installed
	if has gpg-agent; then
		gpg-connect-agent /bye
	fi >&2
}


# Filter rules

rule_match() {
	# Match requires >=6.5
	# Replace with a match-all (Host *) or match-none (default, remove upto the
	# next Match or Host directive) depending on #default: directive
	[ $SSH_MAJOR -gt 6 ] && return
	[ $SSH_MAJOR -eq 6 ] && [ $SSH_MINOR -ge 5 ] && return

	cat <<-'END_AWK'
		BEGIN { is_match=0; match_fallback=false }
		match($1, /^#@fallback:$/) {
		  sub(/.*#@fallback:[[:space:]]*/, "")
		  match_fallback=$0
		  next
		}
		match($1, /^Match$/) {
		  if ( ! system(match_fallback) ) {
		    sub(/^.*$/, "Host *")
		  } else {
		    is_match=1
		  }
		}
		match($1, /^Host$/) { is_match=0 }
		is_match { remove_line() }
		!match($1, /^#fallback:$/) { match_fallback=false }
	END_AWK
}

rule_unix_sockets() {
	# Windows has no unix sockets so disable shared connections
	# (Control* settings)
	if [ $KERNEL = nt ]; then
		echo 'match($1, /^Control/) { print $1 > "/dev/stderr"; remove_line() }'
	fi
}

rule_mux() {
	# ControlMaster & ControlPath require >=3.9
	[ $SSH_MAJOR -gt 3 ] && return
	[ $SSH_MAJOR -eq 3 ] && [ $SSH_MINOR -ge 9 ] && return
	echo 'match($1, /^Control(Master|Path)/) { remove_line() }'
}

rule_mux_persist() {
	# ControlPersist requires >=5.6
	[ $SSH_MAJOR -gt 5 ] && return
	[ $SSH_MAJOR -eq 5 ] && [ $SSH_MINOR -ge 6 ] && return
	echo 'match($1, /^ControlPersist$/) { remove_line() }'
}

rule_canonisation() {
	# OpenSSH < 6.6 does not have name canonisation
	# (Canonical* settings)
	[ $SSH_MAJOR -gt 6 ] && return
	[ $SSH_MAJOR -eq 6 ] && [ $SSH_MINOR -ge 6 ] && return
	echo 'match($1, /^Canonical/) { remove_line() }'
}

rule_percent_C() {
	# %C requires >=6.7
	[ $SSH_MAJOR -gt 6 ] && return
	[ $SSH_MAJOR -eq 6 ] && [ $SSH_MINOR -ge 7 ] && return
	cat <<-'END_AWK'
		match($1, /^(ControlPath|LocalCommand)/) {
			# Temporarily encode %% as the invalid line-feed character to 
			# prevent accidental replacement of "%%C" with "%%l:%r@%h:%p"
			gsub(/%%/, "\n")
			gsub(/%C/, "%l:%r@%h:%p")
			gsub(/\n/, "%%")
		}
	END_AWK
}

rule_include() {
	# Include requires >=7.2
	# If including config.local, remove as it is concatenated by this script
	cat <<-'END_AWK'
		match($1, /^Include$/) {
		  gsub("(~/.ssh/)?config.local", "")
		  if ( NF == 1 ) remove_line()
		}
	END_AWK
	[ $SSH_MAJOR -gt 7 ] && return
	[ $SSH_MAJOR -eq 7 ] && [ $SSH_MINOR -ge 2 ] && return
	echo 'match($1, /^Include$/) { remove_line() }'
}

rule_jump() {
	# ProxyJump requires >=7.3
	# Convert ProxyJump to ProxyCommand
	[ $SSH_MAJOR -gt 7 ] && return
	[ $SSH_MAJOR -eq 7 ] && [ $SSH_MINOR -ge 3 ] && return
	echo 'match($1, /^ProxyJump/) { sub(/^ProxyJump/, "ProxyCommand ssh -W %r@%h:%p") }'
}
